/***************************************************************************
                          Crossfire.h  -  description
                             -------------------
    begin                : Sun May 25 2003
    copyright            : (C) 2003 by Tyler Montbriand
    email                : tsm@accesscomm.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __CROSSFIRE_H__
#define __CROSSFIRE_H__

#define __GAME_CPP__

#include <spritelib/spritelib.h>
#include "StringTable.h"
#include "LoadData.h"
#include "GameGrid.h"
#include "GridSprite.h"
#include "Sample.h"
#include "Game.h"
#include "Music.h"
#include "Score.h"
#include "StringSprite.h"
#include "ScoreSprite.h"

/*! \brief Defines the delay between new bonuses appearing on the board */
#define BONUS_DELAY RandRange(10000,30000)
/*! \brief When only this # of active slots remain, hunt the player down */
#define DEATHHUNT_LEVEL 8

/*! \brief The amount of points scored when the player destroys a monster */
#define POINTS_MONSTER      10
/*! \brief Amount of points scored when player sets off a nuke */
#define POINTS_NUKE         20
/*! \brief Amount of points scored when player takes a freelife bonus */
#define POINTS_LIFE         0
/*! \brief Player will get free lives when multiples of this number scored*/
#define POINTS_FOR_FREELIFE 1000

/*! \brief Sanity-checking..  insane point values can cause event loops */
#if(POINTS_LIFE >= POINTS_FOR_FREELIFE)
#error "Infinite loop condition - lower POINTS_LIFE"
#elif(POINTS_NUKE >= POINTS_FOR_FREELIFE)
#error "Infinite loop condition - lower POINTS_NUKE"
#elif(POINTS_MONSTER >= POINTS_FOR_FREELIFE)
#error "Infinite loop condition - lower POINTS_MONSTER"
#endif

/**
 *  Enumerated type describing different kinds of game sprites
 *  loaded at run-time.
 */
typedef enum GameSprite_e
{
  /*! \brief Player bullets                                   */
  GAME_SPRITE_PFIRE,
  /*! \brief Enemy bullets                                    */
  GAME_SPRITE_EFIRE,
  /*! \brief The player sprite                                */
  GAME_SPRITE_HERO,
  /*! \brief The enemy sprite                                 */
  GAME_SPRITE_MONSTER,
  /*! \brief Simple explosion                                 */
  GAME_SPRITE_BLOWUP,
  /*! \brief Corridor blocks                                  */
  GAME_SPRITE_BLOCK,
  /*! \brief Skull'n'crossbones                               */
  GAME_SPRITE_JOLLY,
  /*! \brief Don't let the radiation in, even if it knocks.   */
  GAME_SPRITE_NUKE,
  /*! \brief Free-life bonus                                  */
  GAME_SPRITE_FREELIFE,
  /*! \brief <voice of doom>Game Over</voice of doom>         */
  GAME_SPRITE_GAMEOVER,
  /*! \brief Title screen                                     */
  GAME_SPRITE_TITLE,
  /*! \brief Game-board background image                      */
  GAME_SPRITE_BACKGROUND,
  /*! \brief Do I really need to explain this?                */
  GAME_SPRITE_PAUSE,
  /*! \brief A simple ASCII bitmap font stored in bitmap form */
  GAME_SPRITE_FONT,
  /*! \brief Sprites >= this index don't exist.               */
  GAME_SPRITE_LAST
} GameSprite_e;

/**
 *  Enumerated type describing the different sounds loaded
 *  at runtime.
 */
typedef enum GameSound_e
{
  /*! \brief Sound made when the player fires.          */
  GAME_SOUND_PFIRE,
  /*! \brief Sound made when the enemy fires.           */
  GAME_SOUND_EFIRE,
  /*! \brief Explosion sound #1.                        */
  GAME_SOUND_EXPLODE,
  /*! \brief Explosion sound #2.                        */
  GAME_SOUND_EXPLODE2,
  /*! \brief Explosion sound #3.                        */
  GAME_SOUND_EXPLODE3,
  /*! \brief The nuke makes a satisfying CRUNCH.        */
  GAME_SOUND_NUKE,
  /*! \brief Sound made when freelife is scored.        */
  GAME_SOUND_FREELIFE,
  /*! \brief The sound the jolly roger makes when shot. */
  GAME_SOUND_LAUGH,
  /*! \brief The sound made when game is paused/upaused */
  GAME_SOUND_PAUSE,
  /*! \brief Sounds >= to this index don't exist.       */
  GAME_SOUND_LAST
} GameSound_e;

/**
 *  Enumerated type describing the collision layers of
 *  various sprites in the game.  Crafted to match the
 *  .spr files.  Should replace this with values read
 *  from sprites themselves, in case the .spr files
 *  change. 
 */
typedef enum GameLayers_e
{
  /*! \brief Collision layer enemies occupy               */
  GAME_LAYER_ENEMY = COL_LAYER(0),
  /*! \brief Collision layer enemy bullets occupy         */
  GAME_LAYER_EFIRE = COL_LAYER(1),
  /*! \brief Collision layer the player occupies          */
  GAME_LAYER_HERO  = COL_LAYER(2),
  /*! \brief Collision layer player bullets occupy        */
  GAME_LAYER_PFIRE = COL_LAYER(3),
  /*! \brief Collision layer the Jolly Roger occupies     */
  GAME_LAYER_JOLLY = COL_LAYER(4),
  /*! \brief Collision layer that nukes occupy            */
  GAME_LAYER_NUKE  = COL_LAYER(5),
  /*! \brief Collision layer that freelife-bonuses occupy */
  GAME_LAYER_LIFE  = COL_LAYER(6)
} GameLayers_e;

/**
 *  Enumerated type describing various global game-events,
 *  communicated to the game object and possibly to sprites
 *  through SDL_UserEvents.
 */
typedef enum GameEvent_e
{
  /*! \brief Event produced when the player dies. */
  /**
   *  Does not occur when the game ends.  Provides no parameters. 
   */
  GAME_EVENT_PLAYERDIED,
  /*! \brief Event produced when the player fires. */
  /**
   *  Used to keep track of the number of shots a player has fired.
   *  Provides no parameters.
   */
  GAME_EVENT_PSHOTSTART,
  /*! \brief Event transmitted when a player shot expires.  */
  /**
   *  Used to keep track of the number of shots a player has fired.
   *  Provides no parameters.
   */
  GAME_EVENT_PSHOTDONE,
  /*! \brief Game event transmitted by enemies, as a request to fire. */
  /**
   *  Parameters are (GAME_EVENT_EFIRE,enemy_sprite_pointer,direction)
   *  See GridSprite.h for the definition of the direction bitmask.
   */
  GAME_EVENT_EFIRE,
  /*! \brief Game event transmitted to request a sound-effect to be played */
  /**
   *  Parameters are (GAME_EVENT_PLAYSFX, GameSound_e, sprite pointer)
   *  or (GAME_EVENT_PLAYSFX, GameSound_e, NULL )
   *
   *  When a sprite is given, the current X position is used to calculate
   *  panning, or when NULL is given, the sound is played at center panning.
   */
  GAME_EVENT_PLAYSFX,
  /*! \brief Game event transmitted to request an explosion */
  /**
   *  The Crossfire engine places an explosion animation at specified
   *  coordinates in response to this message.
   *
   *  Parameters are (GAME_EVENT_EXPLOSION,x,y), where x and y
   *  are signed fixed-point coordinates.
   */  
  GAME_EVENT_EXPLOSION,
  /*! \brief Game event transmitted when a nuke is triggered. */
  /**
   *  In response to this event, a whole bunch of random explosions
   *  are placed on the board, and all enemies that aren't sheltered
   *  in their hiding spots are killed.
   */
  GAME_EVENT_NUKE,
  /*! \brief Game event transmitted when a free life is gained. */
  /**
   *  In response to this event, the freelife sound is played and
   *  the Crossfire::playerlives member is incremented.
   */
  GAME_EVENT_FREELIFE,
  /*! \brief Game event transmitted when the player shoots an enemy. */
  /**
   *  In response to this event, the Crossfire::score member is incremented
   *  POINTS_MONSTER points.
   */
  GAME_EVENT_MONSTERDIED,
  /*! \brief Game event causing enemies to hunt to the death. */
  /**
   *  This event is generated internally by the Crossfire engine and
   *  passed onto all sprites.  When this happens, the enemy sprites
   *  enter a state where, instead of their behavior being determined
   *  randomly, they always move onto the board and hunt down the
   *  player.
   */  
  GAME_EVENT_DEATHHUNT,
  /*! \brief Game event transmitted when the game unpauses. */
  /**
   *  This event is generated internally by the Crossfire engine and
   *  passed onto all sprites when the game is un-paused.
   */   
  GAME_EVENT_UNPAUSE,
  /*! \brief Game event transmitted when an enemy slot is emptied out. */
  /**
   *  When an enemy slot is emptied out entirely, the enemy sprite
   *  transmits this message before removing itself from the board.
   *
   *  The Crossfire engine keeps track of the number of slots that have
   *  been emptied, and when there only DEATHHUNT_LEVEL slots left,
   *  transmits a GAME_EVENT_DEATHHUNT message.
   */
  GAME_EVENT_MONSTERGONE,
  /*! \brief Game event transmitted when the player runs out of lives. */
  /**
   *  When Crossfire::playerlives reaches zero, the Crossfire event
   *  engine transmits this message.  The player sprite removes itself
   *  from the board in response to this message, and a GAME_SPRITE_GAMEOVER
   *  sprite scrolls down the screen.  After a short delay, this sprite
   *  generates a SDL_QUIT message which halts the game engine.
   */    
  GAME_EVENT_GAMEOVER
} GameEvent_e;

class Crossfire:public SpriteManager
{
public:
  /*! \brief Constructor */
  Crossfire();

  /*! \brief Destructor  */
  virtual ~Crossfire();

  /*! \brief Status indicator - returns true when OK, false on error. */
  bool Good();
  /*! \brief Clear()s then Set()s the game board. */
  void Reset();
  /*! \brief Adds game sprites to the board. */
  void Set();
  /*! \brief Clears all sprites from the board. */
  void Clear();

  /*! \brief Timer callback, which calls Crossfire::OnTimer() */
  /**
   *  This is simply a way to translate a static, aka C-style
   *  function call from an SDL timer to a local C++ member
   *  function call.  data is a pointer to the Crossfire
   *  object that created the timer.
   */
  static Uint32 TCallback(Uint32 interval, void *data)
  {
    Crossfire *cfire=(Crossfire *)data;
    return(cfire->OnTimer(interval));  
  }

  /*! \brief Function called in response to the SDL timer */
  Uint32 OnTimer(Uint32 interval);

  /*! \brief Hack to get the titlescreen-image from the Crossfire object */
  /**
   *  I needed a way to get an image out of the Crossfire object.
   *  So sue me.
   */
  VCS_Surface *GetTitle()
  {
    if(Sprite_Loaders==NULL)
    {
      return(NULL);
    }
    else
    {
      return(Sprite_Loaders[GAME_SPRITE_TITLE]->GetSurface());
    }
  }

  /*! \brief Event handler for SDL events that SpriteManager doesn't use. */
  virtual void Handler(SDL_Event *event);
  /*! \brief Function called when SpriteManager has time to spare. */
  virtual void Idle();
  /*! \brief Function handler that SpriteManager calls for SDL_USEREVENTs. */
  virtual int Event(int code, void *data1, void *data2);

  /*! \brief Function called when points are scored. */
  void ScorePoints(int amount);
  /*! \brief Function called when the player attempts to fire. */
  void PlayerFire(int dir);
  /*! \brief Function called when an enemy attempts to fire. */
  void EnemyFire(GridSprite *spr, int dir);  
  /*! \brief Function to attempt to play a sound effect. */
  int PlaySFX(int index, Uint8 dist=0, Sint16 pan=0);
  /*! \brief Places an explosion animation at the specified coordinates. */
  void Explosion(Sfxp x, Sfxp y);
  /*! \brief Causes events and graphics for when a nuke's triggered.*/
  void DropNuke();  

protected:
  /*! \brief Attempts to load and parse the .crossfire configuration file. */
  void Process_Conf();
  /*! \brief Adds the enemy sprites to the board. */
  void Add_Monsters();
  /*! \brief Unloads all game data. */
  void Free_Data();
  /*! \brief Attempts to load all game data. */
  int Init_Data();

  /*! \brief The path where Init_Data() attempts to find graphics files. */
  /**
   *  This variable is read in from .crossfire by Process_Conf()
   */
  char                 Path_Graph[300];
  /*! \brief The path where Init_Data() attempts to find sound files. */
  /**
   *  This variable is read from .crossfire by Process_Conf()
   */
  char                 Path_Sound[300];

  /*! \brief An array of strings, comprising a music list. */
  /**
   *  .crossfire specifies a music list file which is read into
   *  this variable.
   */
  StringTable          Mus_Names;
  /*! \brief An array of strings specifying .spr files to load at runtime. */
  /**
   *  This variable is initialized from a global string array by
   *  Crossfire::Crossfire().  It is my intention to eventually have
   *  these files read in from file as well.
   */
  StringTable          Sprite_Names;
  /*! \brief Array of SpriteLoaders, holding sprite data loaded at runtime  */
  /**
   *  This array corresponds directly to GameSprite_e, that's the array index.
   *  At runtime, this array is allocated and each .spr file specified by
   *  Sprite_Names is loaded in order.
   */
  class SpriteLoader **Sprite_Loaders;
  /*! \brief Array of strings specifying .wav files to load at runtime */
  /**
   *  Also initialized from a global string array by Crossfire::Crossfire().
   *  It is my intention to eventually have these read from file instead
   *  of being statically linked.
   */
  StringTable          Sound_Names;
  /*! \brief Array of sound objects loaded at runtime */
  /**
   *  This array corresponds directly to GameSound_e, that's the array index.
   *  At runtime, this array is allocated and each .wav file specified by
   *  Sound_Names is loaded in order.
   */
  class Sample       **Sound_Data;

  /*! \brief Sound mixer object */
  /**
   *  This is the object that handles all sound-effect playing.
   */  
  class Mixer         *mixer;
  /*! \brief The music player object */
  /**
   *  This is the object that handles the playing of music.
   *  the Music class is a singleton class with a private
   *  constructor;  this means that it must be initialized at
   *  link-time, not run-time, and therefore must be a static
   *  member.
   */
  static class Music   music;
  /*! \brief Player sprite object. */
  /**
   *  Unlike the other sprites, this sprite isn't deleted when it expires
   *  from the game board.  this is because many other objects store a
   *  pointer to it, and deleting it would cause them to reference
   *  a nonexistent object.  As such this object is created before
   *  the game starts and after it ends, not during.
   */
  class Hero          *hero;

  /*! \brief Zero means the player is still alive, nonzero means dead. */
  int                  gameover;
  /*! \brief Number of shots available for the playe to fire. */
  /**
   *  This number is decremented when the player fires and incremented
   *  when the bullets expire.  The player can only fire when this number
   *  is greater than zero.
   */  
  int                  shots;
  
  /*! \brief The number of enemy slots that have been emptied out. */
  /**
   *  The game engine keeps a tally of how many slots have been totally
   *  cleared out in this variable.  When there are only DEATHHUNT_LEVEL
   *  slots left alive, it issues a GAME_EVENT_DEATHHUNT message.
   */
  int                  monsterdeaths;
  /*! \brief Keeps track of scores and high-scores. */
  /**
   *  Keeps a tally of the player's current score, and the overall high-score
   *  from all runs of the game.  Only the current score is cleared by
   *  Clear(), not the high score.  It also keeps track of when the player
   *  has scored enough points to earn another life.
   */
  ScoreSprite          score;
  /*! \brief SDL timer object, that calls TCallBack() */
  SDL_TimerID          timer;
};

/*! \brief The maximum number of shots a player can fire at once. */
#define MAX_PLAYER_SHOTS 2
/*! \brief The speed at which enemy shots travel. */
#define MONSTER_SHOT_SPEED 150
/*! \brief The "distance", aka inverse volume, of enemy shot SFX. */
#define MONSTER_SHOT_DISTANCE 64
/*! \brief The "distance", aka inverse volume, of player shot SFX. */
#define PLAYER_SHOT_DISTANCE  128

/*! \brief The generic "shot" class */
/**
 *  The class that enemy and hero shots are based on.
 *  The only differences from a basic sprite are that
 *  a) They are always deleted when they expire
 *  b) They expire when they leave the boundaries of the game board
 *  c) They expire in response to GAME_EVENT_GAMEOVER and GAME_EVENT_PLAYERDIED
 */
class Shot:public Sprite
{
public:
  /*! \brief Constructor */
  /**
   *  Ensures that the sprite is deleted when it expires.
   */
  Shot()  { SetFlags(SPRITE_DELETEABLE); }

  /*! \brief Clone operator */
  /**
   *  Produces a perfect copy of the sprite, sharing all data,
   *  except the clone doesn't delete any of the shared data
   *  when it expires.
   */
  virtual Sprite *Clone()
  {
    Shot *shot=new Shot();
    
    (*shot)=(*this);
    shot->ClearFlags(SPRITE_OWNSURFACE|SPRITE_OWNBITMASK);
    return(shot);
  }

  /*! \brief Handler for SDL_USEREVENTs */
  /**
   *  Causes shots to expire in response to GAME_EVENT_NUKE
   *  and GAME_EVENT_PLAYERDIED
   */
  virtual void Event(int code, void *data1, void *data2)
  {
    switch(code)
    {
    case GAME_EVENT_NUKE:
    case GAME_EVENT_PLAYERDIED:
      Expire();
      break;

    default:
      break;
    }
  }

  /*! \brief Gives the sprite a starting point, speed, and direction */
  /**
   *  Moves the shot to the location of the source sprite, and moves
   *  away from it at the specified direction and speed.
   */
  void Start(Sprite *source, int speed, int dir)
  {
    if(source!=NULL)  pos.SetFxp(source->Xfxp(),source->Yfxp());

    if(dir&DIR_LEFT)        vel.SetFxp(-speed,0);
    else if(dir&DIR_RIGHT)  vel.SetFxp(speed,0);
    else if(dir&DIR_UP)     vel.SetFxp(0,-speed);
    else if(dir&DIR_DOWN)   vel.SetFxp(0,speed);
    else
    {
      fprintf(stderr,"Shot::Start() - error, no direction given!\n");
      Expire();
    }
  }

  /*! \brief Update tick handler */
  /**
   *  Checks if the shot's gone beyond the game board's bounds, and
   *  expires if it has.
   */
  virtual void Update(Sint32 dtick)
  {
    int x=pos.Xfxp(),y=pos.Yfxp();
    if((x<GMIN_X)||(x>GMAX_X))        Expire();
    else if((y<GMIN_Y)||(y>GMAX_Y))   Expire();
    else                              Sprite::Update(dtick);
  }

  /*! \brief On collision with anything, expire. */
  virtual void OnCollide(const Sprite *spr, ColType index)
  {
    Expire();
  }
};

/*! \brief A simple animation class */
/**
 *  Simply plays all frames in order then expires.
 */
class Animation:public Sprite
{
public:
  /*! \brief Constructor ensures the sprite is deleted after expiration */
  Animation():Sprite()
  { SetFlags(SPRITE_DELETEABLE); }

  virtual void Update(Sint32 dtick)
  {
    /* Current spritelib has "issues" with totally offscreen sprites       */
    if(Xfxp()<GMIN_X)                 Expire();
    else if(Yfxp()<GMIN_Y)            Expire();
    else if(Xfxp()>=(GMAX_X+GRID_X))  Expire();
    else if(Yfxp()>=(GMAX_Y+GRID_Y))  Expire();
    
    Sprite::Update(dtick);
  }

  /*! \brief Overloaded GetFrame operator expires it when it finishes   */
  virtual int GetFrame()
  {
    if(FROM_SFXP(frame)>=(frames.x*frames.y))  Expire();
    return(Sprite::GetFrame());
  }
};

/*! \brief Player shot class */
/**
 *  The only differences from the basic Shot are it
 *  notifying the Crossfire event engine when it's
 *  created and destroyed.
 */
class PShot:public Shot
{
public:
  /*! \brief Notifies game upon creation. */
  PShot()
  { PushUserEvent(GAME_EVENT_PSHOTSTART); }

  /*! \brief Produces perfect clone that shares but doesn't own internal data*/
  virtual Sprite *Clone()
  {
    PShot *pshot=new PShot();
    (*pshot)=(*this);
    pshot->ClearFlags(SPRITE_OWNSURFACE|SPRITE_OWNBITMASK);
    return(pshot);
  }
  /*! \brief Notifies game upon destruction */  
  virtual ~PShot(){ PushUserEvent(GAME_EVENT_PSHOTDONE); }    
};


#endif/*__CROSSFIRE_H__*/
