/***************************************************************************
                          GameGrid.c  -  description
                             -------------------
    begin                : Sun May 11 2003
    copyright            : (C) 2003 by Tyler Montbriand
    email                : tsm@accesscomm.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "GameGrid.h"
#include <stdio.h>
#include <stdlib.h>

/**
 *  An absolute value function that works for integers and
 *  signed fixed point values.
 */
int Abs(int val)
{
  if(val<0) return(-val);
  else      return(val);
}

/**
 *  Calculates the value of y based on the equation
 *
 *    val-min1       x-min2
 *   ----------- = -----------
 *    max1-min1     max2-min2
 *
 *  Or, less mathily, converts a value between min1 and max1 to
 *  a value between min2 and max2.  Like:
 *
 *  Scale(-10, 10,  0, 10, 10) would return 10,
 *  Scale(-10, 10,  0, 10,  0) would return  5, 
 *  Scale(-10, 10,  0, 10,-10) would return  0.
 *
 *  G_CalcPan uses this, to convert the X coordinate value
 *  to a panning value between G_MIN_PAN and G_MAX_PAN.
 */
int Scale(int min1, int max1, int min2, int max2, int val)
{
  int d1=max1-min1;
  int d2=max2-min2;  
    
  val-=min1;
  val=(d2*val)/d1;
  val+=min2;
  
  return(val);      
}

/**
 *  Returns a value between G_MIN_PAN and G_MAX_PAN depending on x.
 *  Values at or smaller than GMIN_X produce G_MIN_PAN, values
 *  greater or equal to GMAX_X produce G_MAX_PAN, anything inbetween
 *  returns a calculated value.
 */
Sint16  G_CalcPan(Sfxp x)
{
  if(x<GOFF_X)      return(G_MIN_PAN);
  else if(x>GMAX_X) return(G_MAX_PAN);
  else
  {
    int val=Scale(GOFF_X,GMAX_X,(360-G_MIN_PAN),(360+G_MAX_PAN),x);
    return(val);
  }
}

Sfxp G_XPos(int xcount)
{
  return((GRID_X*xcount)+GOFF_X);
}

Sfxp G_YPos(int ycount)
{
  return((GRID_Y*ycount)+GOFF_Y);
}

/**
 *  Returns the distance between a X coordinate and a game column.
 *  Used by G_SnapX.
 */
Sfxp G_DistX(Sfxp x, int xcount)
{
  return(G_XPos(xcount)-x);    
}

/**
 *  Returns the distance between a Y coordinate and a game row.
 *  Used by G_SnapY.
 */
Sfxp G_DistY(Sfxp y, int ycount)
{
  return(G_YPos(ycount)-y);
}

/**
 *  Calculates the new x-coordinate when travelling to column
 *  xcount, at velocity 'speed'(in signed fixed point pixels per tick),
 *  given the amount of time dtick(number of ticks).
 *
 *  Moves exactly to the correct X position, correcting for
 *  overshoot.
 */
Sfxp G_SnapX(Sfxp pos, Sfxp speed, int dtick, int xcount)
{
  Sfxp dest=G_XPos(xcount);
  Sfxp d=dest-pos;
  int dist=dtick*speed;

  if(d==0)             return(dest);
  else if(dist>Abs(d)) return(dest);
  else if(d<0)         return(pos-dist);
  else                 return(pos+dist);
}

/**
 *  Calculates the new Y-coordinate when travelling to row
 *  ycount, at velocity 'speed'(in signed fixed point pixels per tick),
 *  given the amount of time dtick(number of ticks).
 *
 *  Moves exactly to the correct Y position, correcting for
 *  overshoot.
 */
Sfxp G_SnapY(Sfxp pos, Sfxp speed, int dtick, int ycount)
{
  Sfxp dest=G_YPos(ycount);
  Sfxp d=dest-pos;
  int dist=dtick*speed;

  if(d==0)             return(dest);
  else if(dist>Abs(d)) return(dest);
  else if(d<0)         return(pos-dist);
  else                 return(pos+dist);
}

/**
 *  Produces random X/Y coordinates on the game grid.
 *
 *  It converts from fixed point to integer, generates the random
 *  numbers, then converts back to fixed point because the difference
 *  between the G_XPos and G_YPos values exceeded RAND_MAX on some
 *  systems.  (cough*WINDOWS*cough)
 */
void G_GetRandPos(Sfxp *x, Sfxp *y)
{
  if((x==NULL)||(y==NULL)) return;

  if(rand()%2)
  {
    *x=G_XPos(RandRange(2,8)&(~1));
    *y=TO_SFXP(RandRange(FROM_SFXP(G_YPos(2)),FROM_SFXP(G_YPos(8))));
  }
  else
  {
    *y=G_YPos(RandRange(2,9)&(~1));
    *x=TO_SFXP(RandRange(FROM_SFXP(G_XPos(2)),FROM_SFXP(G_XPos(9))));
  }
}

/**
 *  Returns a random integer value between min and max, including
 *  negative values.  if abs(min-max) > RAND_MAX, only values
 *  around min will be produced.
 */
Sint32 RandRange(Sint32 min, Sint32 max)
{
  int diff;
  
  if(max<min)
  {
    int val=max;
    max=min;
    min=val;
  }
    
  diff=max-min;
  int val=rand()%diff;
  return(val+min);
}

/**
 *  This is best illustrated by an example..  say:
 *
 *  VALUE_A,VALUE_B,VALUE_C are integer and/or enumerated values
 *  odds[]={1,2,3,-1} and retval[]={VALUE_A,VALUE_B,VALUE_C}
 *
 *  Then there will be a 1/6 chance that VALUE_A will be returned,
 *  a 2/6 chance that VALUE_B will be returned, and a 3/6 chance
 *  that VALUE_C will be returned
 *
 *  This is useful when you have several different
 *  enumerated constants that you want to be selected between
 *  randomly.
 */
int SwitchOdds(int odds[], int retval[])
{
  int total=0;
  int val=-1,n=-1;

  for(n=0; odds[n]!=-1; n++)  total+=odds[n];

  val=rand()%total;

  for(n=0; odds[n]!=-1; n++)
  {
    if(val<odds[n])
    {
      return(retval[n]);
    }
    else            val-=odds[n];
  }

  return(retval[n-1]);
}
