/***************************************************************************
                          Monster.h  -  description
                             -------------------
    begin                : Mon May 19 2003
    copyright            : (C) 2003 by Tyler Montbriand
    email                : tsm@accesscomm.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __MONSTER_H__
#define __MONSTER_H__

#ifndef __cplusplus
#error "Compilation requires C++"
#endif/*__cplusplus*/

#define __GAME_CPP__

#include <stdlib.h>
#include <spritelib/ticktimer.h>
#include "GridSprite.h"
#include "Game.h"

/*! \brief The number of enemies on the grid */
/**
 *  This number is calculated from the size of the game grid.
 *  It works with arbitrary grid sizes, on the condition that
 *  GCOUNT_X and GCOUNT_Y are odd numbers.
 */
#define MAX_MONSTERS  ((GCOUNT_Y+GCOUNT_X)-6)
/*! \brief The speed the enemies travel at. */
#define MONSTER_SPEED 100

/*! \brief The X grid hiding position for every enemy   */
extern int monster_in_x[];
/*! \brief The Y grid hiding position for every enemy   */
extern int monster_in_y[];
/*! \brief The X grid showing position for every enemy  */
extern int monster_out_x[];
/*! \brief The Y grid showing position for every enemy  */
extern int monster_out_y[];
/*! \brief The X grid move-to-board position for every enemy  */
extern int monster_grd_x[];
/*! \brief The Y grid move-to-board position for every enemy  */
extern int monster_grd_y[];
/*! \brief The direction to take potshots in for every monster. */
extern int monster_pshot_dir[];

/*! \brief Calculates the values for all monster starting positions */
void calc_monster_pos();

/*! \brief The minimum amount of time an enemy spends while hiding  */
#define DELAY_HIDE_MIN  1000
/*! \brief The maximum amount of time an enemy spends while hiding  */
#define DELAY_HIDE_MAX  9000

/*! \brief The minimum amount of time an enemy spends while showing */
#define DELAY_SHOW_MIN  100
/*! \brief The maximum amount of time an enemy spends while showing */
#define DELAY_SHOW_MAX  700
/*! \brief The minimum amount of time an enemy spends while taking potshots */
#define DELAY_PSHOT_MIN 100
/*! \brief The maximum amount of time an enemy spends while taking potshots */
#define DELAY_PSHOT_MAX 700

/*! \brief Enumerated type determining an enemy's current behavior pattern */
enum Behavior_e
{
  /*! \brief Hiding  - The enemy stays safe in it's slot                  */
  BHV_HIDE,
  /*! \brief Showing - The enemy moves into vulnerable position           */
  BHV_SHOW,
  /*! \brief Potshot - The enemy is vulnerable, firing into the grid      */
  BHV_POTSHOT,
  /*! \brief ToGrid  - The enemy moves onto the game grid itself          */
  BHV_TOGRID,
  /*! \brief Hunting - The enemy is in the game grid, hunting the player  */
  BHV_HUNT  
};

/*! \brief The enemy sprite class */
/**
 *  Contains the so-called "AI" that controls the enemy's behavior.
 *  It's a very simple AI, really just a state machine that tells
 *  it where to move in response to randomly chosen behaviors.
 *  A timer tells it how long to keep doing one behavior, so when
 *  it expires it switches to a different one.  Here is the state
 *  diagram.  ooh, ASCII art!
 *
 *                     /--------------------------\/           
 *  BHV_HIDE <-> BHV_SHOW <-> BHV_POTSHOT --> BHV_TOGRID --> BHV_HUNT
 *     ^ ^-/       ^-/         ^-/   |
 *     |-----------------------------/
 *
 *  No, ^-/ is not a tortured smiley, it's an arrow pointing from
 *  something back to itself - meaning an enemy could be in the
 *  BHV_POTSHOT state and switch to the BHV_POTSHOT state again
 *  when a new behavior's picked.
 *
 *  Additonally, whenever an enemy collides with anything it "dies"
 *  and is either removed from the game board if it's out of lives,
 *  or immediately reset to the BHV_HIDE state.
 *
 */
class Monster:public GridSprite
{
public:
  /*! \brief Constructor */
  Monster(int idex, GridSprite *target);
  /*! \brief Function called when enemy gets toasted somehow    */
  void Die();
  /*! \brief Routines to perform currently selected behavior    */
  void Think(Sint32 dtick);
  /*! \brief Handles state transitions while in deathhunt mode  */
  void DeathHunt();
  /*! \brief Effects the state transition to hiding */
  void Hide();
  /*! \brief Effects the state transition to Showing  */
  void Show();
  /*! \brief Effects the state transition to taking potshots  */
  void Potshot();
  /*! \brief Effects the state transition to moving onto the game grid */
  void MoveToGrid();
  /*! \brief Effects the state transition to moving to grid or hunting  */
  void Hunt();
  /*! \brief Effects a change to the specified state.  */
  void DoBehavior(Behavior_e beh);
  /*! \brief Tracks toward the specified grid location in 'step' increments */
  void Seek(int gx, int gy, int step=2);
  /*! \brief Moves the enemy home and puts it into the Hiding state */
  void Reset();
  /*! \brief Tracks toward the player sprite  */
  void SeekPlayer();
  /*! \brief Chooses a new behavior */
  void SwitchBehavior();

  /*! \brief Update function called every game tick */
  virtual void Update(Sint32 dtick);
  /*! \brief SDL_USEREVENT handler, for global game events */
  virtual void Event(int code, void *data1, void *data2);
  /*! Function called when enemy collides with things */
  virtual void OnCollide(const Sprite *spr, ColType index);
  /*! \brief Overloaded GetFrame() operator to change the frame draw order  */
  virtual int GetFrame();  

  /*! \brief Returns home X grid coordinate     */  
  inline int HomeX()    { return(monster_in_x[index]);      }
  /*! \brief Returns home Y grid coordinate     */
  inline int HomeY()    { return(monster_in_y[index]);      }
  /*! \brief Returns show X grid coordinate     */
  inline int OutX()     { return(monster_out_x[index]);     }
  /*! \brief Returns show Y grid coordinate     */
  inline int OutY()     { return(monster_out_y[index]);     }
  /*! \brief Returns to-grid X grid coordinate  */
  inline int ToGrdX()   { return(monster_grd_x[index]);     }
  /*! \brief Returns to-grid Y grid coordinate  */  
  inline int ToGrdY()   { return(monster_grd_y[index]);     }
  /*! \brief Returns potshot direction          */
  inline int PshotDir() { return(monster_pshot_dir[index]); }
  /*! \brief Returns enemy's current state  */
  inline Behavior_e Mode() { return(mode); }
  
protected:
  /*! \brief The enemy's target, the player sprite                  */
  static GridSprite *hero;
  /*! \brief Current mode, 0=normal nonzero=death hunt              */
  int         deathhunt;
  /*! \brief Current 'phase', i.e. how many times it's been killed  */
  int         phase;
  /*! \brief Current state/behavior pattern                         */
  Behavior_e  mode;
  /*! \brief Timer clocking how long it's been in one state         */
  ManualTimer timer;
  /*! \brief The amount of time to stay in this state               */
  Sint32      delay;
  /*! \brief Array index for position arrays                        */
  int         index;
};

#endif/*__MONSTER_H__*/
