/***************************************************************************
                          StringTable.h  -  description
                             -------------------
    begin                : Thu Jun 5 2003
    copyright            : (C) 2003 by Tyler Montbriand
    email                : tsm@accesscomm.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __STRINGTABLE_H__
#define __STRINGTABLE_H__

#ifndef __cplusplus
#error "Requires C++ compilation!"
#endif

#include <malloc.h>
#include <string.h>

#define CHUNK_SIZE  5

/*! \brief Removes newline characters from the end of a string  */
inline void StripNewline(char *str)
{
  int pos=-1;
  if(str==NULL) return;

  pos=strlen(str)-1;

  while(pos>=0)
  {
    if((str[pos]=='\n')||(str[pos]=='\r'))  str[pos]='\0';
    else                                    break;

    pos--;
  }  
}

/*! \brief A string-list class  */
class StringTable
{
public:
  /* \brief Constructor that takes an array of strings */
  StringTable(const char *slist[], int listlen=-1)
  {
    int n=-1;
    curpos=0;
    len=0;
    strlist=NULL;

    if(listlen<0)
    {      
      for(n=0; slist[n]!=NULL; n++) Add(slist[n]);
    }
    else
    {
      for(n=0; n<listlen; n++)      Add(slist[n]);
    }    
  }

  /*! \brief Default constructor  */  
  StringTable()
  {
    curpos=0;
    len=0;
    strlist=NULL;   
  }

  /*! \brief Loads in all the non-blank lines from a text file  */
  bool LoadFile(const char *fname)
  {
    bool retval=false;
    FILE *fp=fopen(fname,"r");
    if(fp==NULL)  return(false);    
    retval=LoadFile(fp);
    fclose(fp);
    return(retval);     
  }

  /*! \brief Loads in all the non-blank lines from a file pointer */
  bool LoadFile(FILE *fp)
  {
    char buf[300]={'\0'};

    if(fp==NULL)  return(false);

    if(fgets(buf,299,fp)!=NULL)
    while(!feof(fp))
    {
      StripNewline(buf);      
      Add(buf);
      if(fgets(buf,299,fp)==NULL) break;
    }

    return(true);
  }

  /*! \brief Adds a single non-blank string to the list */
  bool Add(const char *str)
  {
    int size=-1;
    int pos=-1;
    if(str==NULL)           return(false);
    size=strlen(str);
    if(size==0)             return(false);
    pos=FindNewPos();
    if(pos<0)               return(false);
    if(strlist[pos]!=NULL)
    {
      return(false);
    }
    
    strlist[pos]=(char *)malloc(size+1);
    if(strlist[pos]==NULL)  return(false);

    strcpy(strlist[pos],str);
            
    return(true);
  }

  /*! \brief Returns a pointer to the string array  */
  operator const char **()
  {
    return((const char **)strlist);
  }

  /*! \brief Destructor - Frees all data  */    
  ~StringTable()
  {
    int n=-1;
    
    if(strlist!=NULL)
    {
      for(n=0; strlist[n]!=NULL; n++)
      {
        free(strlist[n]);
        strlist[n]=NULL;        
      }
      free(strlist);
      strlist=NULL;
    }
  }

  /*! \brief Returns number of strings in the list  */
  int Size()  { return(curpos); }

private:
  /*! \brief Returns index to place a new string in, or -1 on failure */  
  int FindNewPos()
  {    
    if((strlist==NULL)||(curpos>=(len-1)))
    {
      if(!Extend())   return(false);      
    }
    
    return(curpos++);        
  }

  /*! \brief Allocates CHUNK_SIZE more elements onto the array  */
  bool Extend()
  {
    if(strlist==NULL) return(CreateArray());
    else
    {
      int n=-1;
      char **nlist=NULL;
      int nsize=len+CHUNK_SIZE;
      
      nlist=(char **)realloc(strlist,nsize*sizeof(char *));

      if(nlist==NULL) return(false);
      
      strlist=nlist;
      len=nsize;

      for(n=curpos; n<len; n++) strlist[n]=NULL;
      
      return(true);            
    }            
  }

  /*! \brief Creates the initial array  */
  bool CreateArray()
  {
    int n=-1;
    
    if(strlist!=NULL) return(false);

    strlist=(char **)malloc(sizeof(char *)*CHUNK_SIZE);
    if(strlist==NULL) return(false);
    
    len=CHUNK_SIZE;

    for(n=0; n<CHUNK_SIZE; n++) strlist[n]=NULL;
    
    return(true);
  }
  /*! \brief Actual capacity of the string array  */    
  int len;
  /*! \brief Number of used pointers in the string array  */
  int curpos;
  /*! \brief The string array itself  */  
  char **strlist;    
};

#endif/*__STRINGTABLE_H__*/
